/*
	Copyright 2012, Marcel Lambrechts <marcel.lambrechts@gmail.com>

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
	
	Created: 05-10-2012
	Version: 1.0
	
	Command line arguments:
	sworapwd -p <oracle password file>
	
	Decription:
	This utility will let you switch the specified password file between
	case-sensitive and case-insensitve mode for 11g (11.1 and 11.2)
	databases and ASM instances. Using this utility you won't have to
	recreate a complete password file and recreate password file users
	(and grant permissions) if you would like to switch.
	
	This utility will take the name of an password file. First it will create
	a backup of the password file it will be changing in a file with the same
	filename (path) and the date/time added to it.
	
	It will check the "endianess" of the password file, do some check to make
	sure the file is actually an Oracle password file and then it will switch
	the "ignorecase" bit for the specified password file.
*/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "getopt.h"

#define TBSIZE 256

int main(int argc, char* argv[])
{
	FILE *pfp, *bfp;
	char *buffer;
	unsigned char cases;
	int fnlength = 0;
	unsigned long filesize = 0;
	char *pfilemagic = "ORACLE Remote Password file";
	time_t curtime;
	struct tm *loctime;
	char timebuffer[TBSIZE];
	
	// Command-line argument
	char c;
	int args = 0;
	char pwdfile[256];
	char bckfile[272];					// 256 + 16 characters added to passwordfile name for backup
	
	// Endianess detection and case-sensitivity detecten
	int endianess = 0;					// 0 = little endian, 1 = big endian
	int bufindex = 0;
	unsigned char bemagic[4] = { 0x5A, 0x5B, 0x5C, 0x5D };
	unsigned char lemagic[4] = { 0x5D, 0x5C, 0x5B, 0x5A };
	
	// Get Command-line options
	while ((c = getopt(argc, argv, "p:")) != -1){
		switch (c) {
		case 'p':
			if (strlen(optarg) > 256) {
				printf("Specified passwordfile name cannot exceed 256 characters\n");
				return -1;
			}
			fnlength = strlen(optarg);
			memcpy(pwdfile, optarg, fnlength + 1);
			args++;
			break;
		}
	}
	
	// Make sure both command line options are specified
	if(args!=1){
		printf("%s -p <oracle password file>\n", argv[0]);
		return -1;
	}
	
	// Generate a name for the backup of the password file
	memcpy(bckfile, pwdfile, strlen(pwdfile) + 1);
	curtime = time (NULL);
	loctime = localtime (&curtime);
	strftime (timebuffer, TBSIZE, "_%d%m%Y_%H%M%S", loctime);
	strcat(bckfile, timebuffer);
	printf("[INFO] Backup password file: %s\n", bckfile);
	
	// Open the password file (read/write)
	pfp = fopen(pwdfile, "r+b");
	if (!pfp)
	{
		fprintf(stderr, "Unable to open file %s", pwdfile);
		return 1;
	}
	
	// Determine length of password file
	fseek(pfp, 0, SEEK_END);
	filesize=ftell(pfp);
	fseek(pfp, 0, SEEK_SET);

	// Allocate memory to store contents of password file
	buffer=(char *)malloc(filesize+1);
	if (!buffer)
	{
		fprintf(stderr, "Memory error!");
		fclose(pfp);
		free(buffer);
		return 1;
	}

	// Read password file in memory
	fread(buffer, filesize, 1, pfp);
	
	// Make sure we are working on a password file
	if (memcmp(buffer+0x200, pfilemagic,27)!=0) {
		fprintf(stderr, "Not an Oracle password file\n");
		fclose(pfp);
		free(buffer);
		return 1;
	}
	
	// Check endianess of password file
	if (memcmp(buffer+0xC, bemagic, 4)==0) {
		endianess = 1;
		printf("[INFO] Big Endian password file found\n");
	} else if (memcmp(buffer+0xC, lemagic, 4)==0) {
		endianess = 0;
		printf("[INFO] Little Endian password file found\n");
	} else {
		fprintf(stderr, "Not an (supported) Oracle password file\n");
		fclose(pfp);
		free(buffer);
		return 1;
	}
	
	// Open backup password file for writing (truncating)
	bfp = fopen(bckfile, "wb");
	if (!pfp)
	{
		fprintf(stderr, "Unable to open backup file %s", bckfile);
		fclose(pfp);
		free(buffer);
		return 1;
	}
	
	// Write buffer (copy) to backup password file and close it
	if (fwrite(buffer, filesize, 1, bfp) != 1) {
		fprintf(stderr, "Error writing to backup password file. Terminating %s\n", argv[0]);
		fclose(pfp);
		fclose(bfp);
		free(buffer);
		return 1;
	}
	fclose(bfp);
	
	// byte containing base-sensitivity bits depend on endianess of file (originally 2 byte value)
	if (endianess==0) {
		// Little Endian
		bufindex = 0x240;
	} else {
		// Big Endian
		bufindex = 0x241;
	}
	
	// See what we should switch to (ignorecase=n or ignorecase=y)
	if ((buffer[bufindex]&2)==0) {
		printf("[INFO] switch to case-sensitive (ignorecase=n)\n");
	} else {
		printf("[INFO] switch to case-insensitive (ignorecase=y)\n");
	}
	fseek(pfp, bufindex, SEEK_SET);
	cases = buffer[bufindex]^2;
	fwrite(&cases, sizeof(cases), 1, pfp);

	// Close passwordfile handle, free buffer memory and exit
	fclose(pfp);
	free(buffer);
	return 0;
}
